function [D,X] = DL_mask_multi(Y, Ymask, Ypos, D, d_size, s, iternum)

% Input:
%   Y           - training data
%   Ymask       - mask of missing data (1-available, 0-missing)
%   Ypos        - start positions in Y of data belonging to different series
%                 nseries = length(Ypos)!
%   D           - initial dictionary (must have sum(d_size) atoms)
%   d_size      - vector of dictionary sizes:
%                   d_size(i) - size of dictionary for series i, i=1:nseries
%                   d_size(nseries+1) - size of common dictionary
%   s           - sparsity level
% Output:
%   D           - designed dictionary, structured as stated by d_size
%   X           - representations for final dictionary

% BD 15.12.2019

[m,n_tot] = size(D);
N = size(Y,2);
if n_tot ~= sum(d_size)
  error('Number of atoms in D0 must be sum(d_size)')
end
nseries = length(Ypos);     % number of data series
if length(d_size) ~= nseries+1
  error('length(d_size) must be length(Ypos)+1') 
end
if size(Y) ~= size(Ymask)
  error('Y and Ymask must have the same size')
end
d_size = d_size(:)';

err = zeros(1,iternum);
Ypos(nseries+1) = size(Y,2)+1;  % for easier manipulation
dpos = 1 + [0 cumsum(d_size)];  % start positions for individual dictionaries
idcomm = dpos(end-1):dpos(end)-1;     % positions of common atoms

% main DL loop
for iter = 1 : iternum
  % sparse representation, separately for each series
  X = zeros(n_tot,N);   % size of overall representations
  for is = 1 : nseries  % representations for each series of signals
    ids = dpos(is):dpos(is+1)-1;
    Ds = [D(:,ids) D(:,idcomm)]; % dictionary: individual and common parts
    iys = Ypos(is):Ypos(is+1)-1;
    Xs = omp_mask(Y(:,iys), Ds, s, [], Ymask(:,iys));
    X(ids,iys) = Xs(1:d_size(is),:);   % representations using individual atoms
    X(idcomm,iys) = Xs(d_size(is)+1:end,:);   % representations using common atoms
  end
  
  % atom update using AK-SVD, made in the standard way
  for j = 1:n_tot     
    [~, data_indices, x] = find(X(j,:));
    d = D(:,j);
    D(:,j) = 0;
    if (isempty(data_indices)) % unused atoms are replaced with random vectors
      atom = randn(m,1);
      D(:,j) = atom / norm(atom);
      continue;
    end
    smallX = X(:,data_indices);
    smallY = Y(:,data_indices);
    smallYmask = Ymask(:,data_indices);
    [D(:,j),X(j,data_indices)] = aksvd_mask_up_here(smallY,D,smallX,x,smallYmask);
  end

  % error
  err(iter) = norm(Ymask.*(Y - D*X), 'fro') / sqrt(sum(sum(Ymask)));
end
end

function [d,x] = aksvd_mask_up_here(Y,D,X,x,Ymask)
%% Approximate K-SVD atom update
% INPUTS:
%   Y -- training signals set
%   D -- current dictionary
%   X -- sparse representations
%   x -- sparse representations row using the current atom
%
% OUTPUTS:
%   d -- updated atom
%   x -- updated representations corresponding to the current atom
%     d = Y *x' - D*(X*x');
%     d = d/norm(d);
%     x = Y'*d - (D*X)'*d;
    d = Ymask.*(Y - D*X)*x';
    d = d/norm(d);
    x = (Ymask.*(Y - D*X))'*d;
    for j = 1 : length(x)
      x(j) = x(j) / ((Ymask(:,j).*d)'*(Ymask(:,j).*d));
    end
end
