function run_testing_missing_data(no_runs, perc_miss, flag_polya, par_polya, all_place, flag_transf, flag_plots)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Input:

% no_runs           = no. of initializations for the dictionary 
% perc_miss         = percentage missing data
% flag_polya        = if 1, use Polya urn for positions missing data
%                     if 0, use unif. distrib. for positions missing data 
% par_polya         = vector with two entries:
%                     1st entry: M-"memory" Polya urn
%                     2nd entry: delta-if large, there are bursts of 
%                     missing data instead of isolated missing data
% all_place         = list of times series that are used 
% flag_transf       = if 1, data are transformed; if 0, data are not
%                      transformed
% vflag             = [flag_log flag_lin_trend flag_periodic_comp flag_season]
% vperiod           = periods for the periodic components (used only if flag_periodic_comp~0)
% vseason           = seasonality parameters (used only if flag_season~=0)
% flag_plots        = if 0, does not show the plots
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

close all

if (perc_miss<=0) || (perc_miss>=50/100)
    fprintf('ERROR: perc_miss should be selected from the interval (0,1/2)!\n');
    return;
end

% Name of the data file (without .mat)
fname_c = 'Australia_Tourism';

%Sizes and parameters
signal_shift = 1;               % how many sampling periods from the beginning of a signal to the next
if flag_transf==1
    parameters_transf;          % param. for transforming the time series
else
    parameters_non_transf;      %param. for non-transforming the time series
end
no_it_crt = 2;                  % no. of the IT criteria that are evaluated
iters = 50;                     % number of DL iterations
m = 12*1;                       % signal length, normally a multiple of 24 (we work with Hourly data)
nv = 12*(2:8);                  % dictionary sizes(choosen to be the multiplication of signal_length) 
sv = [2 3 4];                 % sparsity levels


%File for saving the results
fname_out    = strcat('Outputs_perc_', num2str(round(perc_miss*100)), '_place_');    % file in which the results are saved
for k=1:length(all_place)
     fname_out = strcat(fname_out,num2str(all_place(k)),'_');
end
if flag_transf==1
    fname_out = strcat(fname_out,'transformed_data');
else
    fname_out = strcat(fname_out,'original_data');
end
fname_out = strcat(fname_out,'.mat');

%Compute qv
if sum(rem(nv,m))~=0
    fprintf('ERROR: n should be a multiple integer of m!');
    return;
else
    qv = nv./m;
end

%Init err_cell_miss
err_cell_miss = cell(length(nv),1);
for a=1:length(nv)
    err_cell_miss{a,1} = zeros(no_runs, (qv(a)-1), length(sv)); %errors in the estimation of the missing data
end

%Init err_cell_miss_indiv
err_cell_miss_indiv = cell(length(nv),1);
for a=1:length(nv)
    err_cell_miss_indiv{a,1} = zeros(length(all_place), no_runs, (qv(a)-1), length(sv)); %errors in the estimation of the missing data (for each TS)
end

%Init cell_crit
cell_crit = cell(length(nv),1);
for a=1:length(nv)
    cell_crit{a,1} = Inf*ones(no_it_crt, no_runs, (qv(a)-1), length(sv)); %IT criteria
end

%Save init.
if isfile(fname_out)==1
    delete(fname_out);
end
save(fname_out, 'err_cell_miss', 'err_cell_miss_indiv', 'cell_crit');  

%Run test
fprintf('\n');
for crun=1:no_runs
    test_missing_data_multi(fname_c, no_it_crt, iters, m, nv, sv, qv, crun, ...
        perc_miss, flag_polya, par_polya, all_place, signal_shift, ...
        all_vflag, all_vperiod, all_vseason, flag_plots, fname_out);
end
fprintf('\n');

load(fname_out, 'err_cell_miss', 'err_cell_miss_indiv', 'cell_crit');  
best_n          = zeros(no_it_crt+1,1);
best_s          = zeros(no_it_crt+1,1);
best_q          = zeros(no_it_crt+1,1);
min_itc         = Inf*ones(no_it_crt+1,1);
min_err         = Inf*ones(no_it_crt+1,1);
min_err_indiv   = Inf*ones(no_it_crt+1,length(all_place)); 
for a=1:length(nv)
    temp            = cell_crit{a,1};
    temp_err        = err_cell_miss{a,1};
    temp_err_indiv  = err_cell_miss_indiv{a,1};
    qqv             = (1:qv(a)-1);
    for cc=1:no_it_crt
        temp2 = squeeze(temp(cc,:,:,:));
        [mitc,ind] = min(temp2(:));
        if mitc<min_itc(cc)
            min_itc(cc)         = mitc;
            best_n(cc)          = nv(a);
            [crun,aa,b]         = ind2sub(size(temp2),ind);
            best_q(cc)          = qqv(aa)*m;
            best_s(cc)          = sv(b);
            min_err(cc)         = temp_err(crun,aa,b);
            min_err_indiv(cc,:) = squeeze(temp_err_indiv(:,crun,aa,b));
        end %if merr
    end %for cc
    [merr,ind] = min(temp_err(:));
    if merr<min_err(end)
        min_err(end)            = merr;
        best_n(end)             = nv(a);
        [crun,aa,b]             = ind2sub(size(temp_err),ind);
        best_q(end)             = qqv(aa)*m;
        best_s(end)             = sv(b);
        min_err_indiv(end,:)    = squeeze(temp_err_indiv(:,crun,aa,b));
    end
end %for a
fprintf('Results IT criteria:\n');
for cc=1:no_it_crt
    fprintf('ITC%i:\n Av_err=%5.4f\n n=%i\n  s=%i\n Size dict. common=%i\n', ...
        cc, min_err(cc), best_n(cc), best_s(cc), best_n(cc)-best_q(cc) );
    for ts=1:length(all_place)
        fprintf('Err TS%i: %5.4f\n', all_place(ts), min_err_indiv(cc,ts) );
    end
end
fprintf('Results Oracle:\n');
fprintf('Av_err=%5.4f\n n=%i\n  s=%i\n Size dict. common=%i\n', ...
        min_err(end), best_n(end), best_s(end), best_n(end)-best_q(end) );
for ts=1:length(all_place)
    fprintf('Err TS%i: %5.4f\n', all_place(ts), min_err_indiv(end,ts) );
end

end % function

