% Copyright (c) 2016 Paul Irofti <paul@irofti.net>
% 
% Permission to use, copy, modify, and/or distribute this software for any
% purpose with or without fee is hereby granted, provided that the above
% copyright notice and this permission notice appear in all copies.
% 
% THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
% WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
% MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
% ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
% WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
% ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
% OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

function [D,X,shared] = aksvd_size(Y,D,X,iter,replatoms,shared,varargin)
%% AK-SVD algorithm for size-optimal DL
% INPUTS:
%   Y -- training signals set
%   D -- current dictionary
%   X -- sparse representations
%
% PARAMETERS:
%   w1_coef -- promote matrix sparsity for X (default: 6)
%   w1_row -- promote row sparsity for X (default: 6)
%   w1_row_factor -- multiply at each iter with w1_row (default: 2)
%   atomstop -- cancel w1_coef when there are less atoms than atomstop;
%               this uses backtracking when the dictionary size gets below
%               atomstop (default: 0)
%   w1_row_nochange -- cancel w1_row if the size does not change in
%                       w1_row_nochange iterations (default: 0)
%
% OUTPUTS:
%   D -- updated dictionary
%   X -- updated representations
    persistent w1_coef;
    persistent w1_row;
    persistent atomstop;
    % persistent w1_coef_factor;
    % persistent w1_coef_nochange;
    persistent w1_row_factor;
    persistent w1_row_nochange;
    persistent nochange_count;

    if iter == 1
        p = inputParser();
        p.KeepUnmatched=true;
        p.addParameter('w1_coef', 6);
        %p.addParameter('w1_coef_factor', 1);
        p.addParameter('w1_row', 6);
        p.addParameter('w1_row_factor', 1);
        p.addParameter('atomstop', 0, @(x) x <= size(D,2));
        %p.addParameter('w1_coef_nochange', 0);
        p.addParameter('w1_row_nochange', 0);
        p.parse(varargin{:});
        w1_coef = p.Results.w1_coef;
        w1_row = p.Results.w1_row;
        %w1_coef_factor = p.Results.w1_coef_factor;
        w1_row_factor = p.Results.w1_row_factor;
        atomstop = p.Results.atomstop;
        %w1_coef_nochange = p.Results.w1_coef_nochange;
        w1_row_nochange = p.Results.w1_row_nochange;
        nochange_count = 0;
    end

    % Check the atomstop target.
    if atomstop && w1_row ~= 0
        % If we're close enough, stop shrkinking.
        if size(D,2) >= atomstop && size(D,2) < atomstop + 4
            w1_row = 0;
        % Otherwise save the initial dictionary and representations for
        % w1_row manipulations.
        else
            D0 = D;
            X0 = X;
        end
    end
    
    % Save the initial dictioanry and representations for checking
    % if the dictionary shrinked later on.
    if w1_row_nochange && w1_row ~= 0
        D0 = D;
        X0 = X;
    end

    [D,X] = size_atom_up(Y,D,X,replatoms, ...
        @(F,d,x) size_up(F,x,w1_coef,w1_row));
    
    % Change w1_row depending on how close we got to atomstop.
    if atomstop ~= 0 && w1_row ~= 0
        if size(D,2) > atomstop
            % We can still shrink the dictionary.
            w1_row = w1_row * w1_row_factor;
        elseif size(D,2) < atomstop
            % If we went past atomstop, try and recover.
            fprintf('!');
            [D,X,w1_row] = w1_row_retry(Y,D0,X0,replatoms, ...
                w1_coef,w1_row,w1_row_factor,atomstop);
        end
    end
    
    % Check if the dictionary size changed in the last iterations.
    % If it did not, cancel w1_row.
    if w1_row_nochange && w1_row ~= 0
        if size(D0,2) == size(D,2)
            nochange_count = nochange_count + 1;
            % if nochange_count == w1_coef_nochange
            if nochange_count == w1_row_nochange
                % w1_coef = 0;
                % w1_coef_nochange = 0;
                w1_row = 0;
                w1_row_nochange = 0;
            end
        else
            nochange_count = 0;
        end
    end
 
end

function [d,x] = size_up(F,x,w1_coef,w1_row)
%% AK-SVD for size-optimal DL
% INPUTS:
%   Y -- training signals set
%   D -- current dictionary
%   X -- sparse representations
%   d -- current atom
%   w1_coef -- promote matrix sparsity for X
%   w1_row -- promote row sparsity for X
%
% OUTPUTS:
%   d -- updated atom
%   x -- updated representations corresponding to the current atom

	d = F*x';
    d = d/norm(d);
    
    z = sign(F'*d).*max(0, abs(F'*d) - w1_coef*ones(size(x')));
    if z == 0
        x = 0;
    else
        x = 1/norm(z) * max(0, norm(z) - w1_row) * z;
        x = x';
    end
    
    if x == 0
        x = [];
        d = [];
    end
end

function [D,X, w1_row] = w1_row_retry(Y,D,X,replatoms, ...
    w1_coef, w1_row, w1_row_factor, atomstop)
%% Retry size reduction with a new w1_row
% This function is used for backtracking when there are less atoms
% in the new dictionary than stated in atomstop.
%
% w1_row is lowered and a new dictionary is built.
%
% It was needed because recursion does not save the last succesful
% value of w1_row because it is a persistent variable.
%
% INPUTS:
%   Y -- training signals set
%   D -- current dictionary
%   X -- sparse representations
%   replatom -- atom replacement strategy
%   w1_coef -- matrix sparsity for X
%   w1_row -- row sparsity for X
%   w1_row_factor -- change in w1_row
%   atomstop -- do not let the dictionary be smaller than this
%
% OUTPUTS:
%   D -- updated dictionary
%   X -- updated representations
%   w1_row -- a better w1_row

    D0 = D;
    X0 = X;
    first = true;
    maxiter = 10;
    i = 1;
    
    w1down = w1_row/w1_row_factor;  % previous w1_row value
    while (size(D,2) < atomstop && i < maxiter) || first
        w1_row = (w1down + w1_row) / 2;
        [D,X] = size_atom_up(Y,D0,X0,replatoms, ...
            @(F,d,x) size_up(F,x,w1_coef,w1_row));
        
        first = false;
        i = i + 1;
    end
 
    % If shrinking w1_row to get above atomstop failed, cancel w1_row
    % and return the last known dictionary and representations that
    % satisfied atomstop
    if i == maxiter
        D = D0;
        X = X0;
        w1_row = 0;
    end
end
