function [W, D, A] = clas_labelcon_dl(Y, H, Q, n, s, alpha, beta, init_method)
% Design label consistent DL classifier (Section 8.5.3)
% Input:
%   Y           - signal matrix
%   H           - label matrix
%   Q           - label consistency matrix
%   n           - number of atoms
%   s           - sparsity level
%   alpha       -
%   beta        - trade-off parameters in the objective
%   init_method - 0, random
%                 1, trained dictionaries for each class and LS classifier and A
% Output:
%   W           - classifier matrix
%   D           - dictionary for sparse representation
%   A           - label consistency transformation

% BD 22.04.2018

if nargin < 8
  init_method = 0;
end

Ye = [Y; sqrt(alpha)*H; sqrt(beta)*Q];    % extended signal matrix
iters = 50;                 % number of iterations in DL algorithm
iterc = 20;                 % number of iterations for training initial dictionaries
update = 'aksvd';           % dictionary update method
replatom = 'random';        % atom replacement strategy
params = {};
[m,N] = size(Y);            % m=signal length, N=number of training signals
c = size(H,1);              % number of classes
me = m+c+n;

% Label consistent DL
switch init_method
  case 0                    % plain initialization
    D0 = randn(me,n);
    D0 = normc(D0);
  case 1                    % train small dictionaries for each class
    D0 = zeros(me,n);
    jj = 0;
    for i = 1 : c           % train the dictionaries
      jc = find(H(i,:)==1); % indices of signals from class i
      Yc = Y(:, jc);        % those signals
      nc = sum(Q(:,jc(1))); % number of atoms for class i
      D0c = randn(m,nc);
      D0c = normc(D0c);
      Dc = DL(Yc, D0c, s, iterc, str2func(update), params, 'replatom', replatom);
      D0(1:m,jj+1:jj+nc) = Dc;
      jj = jj + nc;
   end
    X = omp(Y, D0(1:m,:), s);
    % compute initial classifier and label consistency transformation
    gamma = 1e-3;
    W = H*X' / (X*X' + gamma*eye(n));
    A = Q*X' / (X*X' + gamma*eye(n));
    D0(m+1:m+c,:) = sqrt(alpha) * W;
    D0(m+c+1:me,:) = sqrt(beta) * A;
    D0 = normc(D0);
end

De = DL(Ye, D0, s, iters, str2func(update), params, 'replatom', replatom);

% Extract dictionary, classifier matrix and label consistency transformation
D = De(1:m,:);
normD = sqrt(sum(D.*D));
D = normc(D);
W = De(m+1:m+c,:) / sqrt(alpha);
W = W ./ repmat(normD, c, 1);
A = De(m+c+1:me,:) / sqrt(beta);
A = A ./ repmat(normD, n, 1);

